#include <stdint.h>
#include <stdio.h>
#include <tgmath.h>
#include <timers.h>
#include <timing.h>
#include <drv_rtc.h>

#include "swplatform.h"

// #define PI 3.14159
#define PI 2.0*asin(1.0)
#define DAC_CHAN_X 0
#define DAC_CHAN_Y 1
#define MINX 0
#define MINY 0
#define MAXX 255
#define MAXY 255

#define MPEREG_ADR_LOW     1
#define MPEREG_ADR_HIGH    2
#define MPEREG_ADR_DPB     3
#define MPEREG_SLOTLENGTH  4
#define MPEREG_RUN 15

// debug/logging levels

#define DBG_NONE  0
#define DBG_FATAL 1
#define DBG_ERROR 2
#define DBG_WARN  3
#define DBG_INFO  4
#define DBG_DEBUG 5

// number of samples to store in memory buffer

#define MAXCIRCLE 360                   // a circle is divided up into 360 steps
#define DEF_BUFSIZE 2000                // 1000 points (1x+1y)
#define DEF_DPBSIZE 41                  // default DPB has 41 entries for 40 descriptor pointers plus NULL end tag

// operational modes

#define OPM_CLOCK_DSC 128
#define OPM_CLOCK   129
#define OPM_WBCTEST 130
#define OPM_DSCTEST 131
#define OPM_CSQTEST 132
#define OPM_MEMTEST 133

#define OPERMODE OPM_CSQTEST

typedef int dim_t;

struct Coord_t {
    float x;
    float y;
};

//
// =============================== Global Variables ==========================================
//

int debug_level = 0;

//
// ================================ auxiliary functions =======================================
//

// set debug level
void setDebug(int lvl) {
    debug_level = lvl;
}

// check debug level
_Bool isDebug(int lvl) {
    return (debug_level>=lvl);
}

// fatal program end
void fatal() {
    printf("FATAL program abort\r\n");
    for (;;) {
    }
}

// convert rad to degrees
float rad2deg(float rad) {
    return (rad*360.0)/(2.0*PI);
}

//
// ================== class DSC - MPE Buffer DeSCriptor ========================
//
struct MPEBufDesc_t {
    uint32_t low_adr;
    uint32_t high_adr;
    uint32_t ext_length;
    uint32_t offset_longword;
    uint32_t mult_longword;
};

typedef struct dbuf_ts {
    uint8_t *byte_buffer;
    int buffer_size;
    int current_write_index;
    struct MPEBufDesc_t dsc;
} dbuf_t;

// allocate a new sample buffer, default buffer size is DEF_BUFSIZE
uint8_t* new_cbuf() {
    void* tt = malloc(DEF_BUFSIZE);
    if (tt<=0) {
        printf("FATAL: Cannot allocate buffer of size %d\r\n",DEF_BUFSIZE);
        fatal();
    }
    return (uint8_t*) tt;
}

// release a buffer
void release_cbuf(uint8_t* bufp) {
    free((void*)bufp);
}

// allocate a DBUF with a byte buffer already provided
dbuf_t * new_dbuf_from_cbuf(uint8_t* cbufp) {
    void* tt = malloc(sizeof(dbuf_t));
    if (tt<=0) {
        printf("FATAL: could not allocate DBUF\r\n");
        fatal();
    }
    dbuf_t* bp = (dbuf_t*)tt;
    bp->byte_buffer = cbufp;
    bp->buffer_size = DEF_BUFSIZE;
    bp->current_write_index = 0;
    if (bp->byte_buffer) {
      bp->dsc.low_adr = (uint32_t)&bp->byte_buffer[0];
      bp->dsc.high_adr = (uint32_t)&bp->byte_buffer[bp->current_write_index];
    } else {
      bp->dsc.low_adr  = 0;
      bp->dsc.high_adr = 0;
    }
    bp->dsc.ext_length      = 2;
    bp->dsc.mult_longword   = 0x00000101;
    bp->dsc.offset_longword = 0;
    return bp;
}

// allocate a DBUF with a default size byte buffer
dbuf_t * new_dbuf() {
    return new_dbuf_from_cbuf(new_cbuf());
}

// allocate a DBUF with NO byte buffer
dbuf_t * new_dbuf_no_cbuf() {
    return new_dbuf_from_cbuf((uint8_t*)0);
}

// set byte buffer addresses
void dbuf_set_cbufadr(dbuf_t *bp,uint32_t cbuflo, uint32_t cbufhi) {
  bp->dsc.low_adr = cbuflo;
  bp->dsc.high_adr = cbufhi;
}

// set DBUF offset values
void dbuf_set_offset(dbuf_t *bp, dim_t ox, dim_t oy) {
    bp->dsc.offset_longword = (oy%256)*256 + (ox%256);
}

// set DBUF multiplier values
void dbuf_set_mult(dbuf_t *bp, dim_t mx, dim_t my) {
    bp->dsc.mult_longword = (my%256)*256 + (mx%256);
}

// draw a point to a buffer
void draw_to_dbuf(dbuf_t *bp, dim_t x, dim_t y) {
    if (bp->buffer_size<=(bp->current_write_index+2)) {
        printf("FATAL: could not write any more sample to buffer\r\n");
        fatal();
    }
    bp->byte_buffer[bp->current_write_index++] = (uint8_t)(x>255?255:x);
    bp->byte_buffer[bp->current_write_index++] = (uint8_t)(y>255?255:y);
    bp->dsc.high_adr = ((uint32_t)&bp->byte_buffer[bp->current_write_index-2]) & 0xfffffffc;
}

void release_dbuf(dbuf_t *bp) {
    release_cbuf(bp->byte_buffer);
    free((void*)bp);
}

uint32_t dbuf_get_dscadr(dbuf_t *bp) {
    return (uint32_t)&bp->dsc;
}

//
// transform coordinates of point <p> according to transformation matrix <a,c,e,b,d,f,0,0,1>
// this is acheived by matrix multiplication
// see also: https://developer.mozilla.org/de/docs/Web/SVG/Attribute/transform
//
struct Coord_t matrix(struct Coord_t p, float a, float b, float c, float d, float e, float f) {
    struct Coord_t res;
    res.x = a*p.x + c*p.y + e;
    res.y = b*p.x + d*p.y + f;
    return res;
}

//
// translate point <p> by <dx> in X- and by <dy> in Y-direction
//
struct Coord_t translate(struct Coord_t p, float dx, float dy) {
    struct Coord_t res = matrix(p,1.0,0.0,0.0,1.0,dx,dy);
    if (isDebug(DBG_DEBUG)) {
      printf("translated point <%5.2f/%5.2f> by <%5.2f/%5.2f> to <%5.2f/%5.2f>.\r\n",p.x,p.y,dx,dy,res.x,res.y);
    }
    return res;
}

//
// scale point <p> by <fx> in X-direction and <fy> in Y-direction
//
struct Coord_t scale(struct Coord_t p, float fx, float fy) {
    struct Coord_t res = matrix(p,fx,0.0,0.0,fy,0.0,0.0);
    if (isDebug(DBG_DEBUG)) {
      printf("scaled point <%5.2f/%5.2f> by <%5.2f/%5.2f> to <%5.2f/%5.2f>.\r\n",p.x,p.y,fx,fy,res.x,res.y);
    }
    return res;
}

//
// rotate point <p> by angle <phi> around <0,0>
//
struct Coord_t rotate(struct Coord_t p, float phi) {
    struct Coord_t res = matrix(p,cos(phi),0.0-sin(phi),0,sin(phi),cos(phi),0);
    if (isDebug(DBG_DEBUG)) {
      printf("rotated point <%5.2f/%5.2f> by %5.2f to <%5.2f/%5.2f>.\r\n",p.x,p.y,rad2deg(phi),res.x,res.y);
    }
    return res;
}

void hexdump_line(uint8_t *bp, uint32_t len, uint32_t pos, uint32_t siz) {
    printf("%04x ",pos);
    int i;
    uint8_t b;
    for (i = 0; i<siz; i++) {
        if ((pos+i)<len) {
           printf("%02x ",*(bp+pos+i));
        } else {
           printf("   ");
        }
    }
    printf(" : ");
    for (i = 0; i<siz; i++) {
        if ((pos+i)<len) {
           b = *(bp+pos+i);
           printf("%c",((b>=32) && (b<127))?(char)b:'.');
        } else {
            printf(" ");
        }
    }
    printf("\r\n");
}

void hexdump(uint8_t *bp, uint32_t len, uint32_t siz) {
  int pos = 0;
  while (pos<len) {
    hexdump_line(bp,len,pos,siz);
    pos += siz;
  }
}

void dump_dsc(uint32_t dscadr) {
    uint32_t ladr =  *((uint32_t*)(dscadr+0));
    uint32_t hadr =  *((uint32_t*)(dscadr+4));
    uint32_t len = hadr-ladr+4;
    printf("DSC(%08x..%08x):\r\n",ladr,hadr);
    hexdump((uint8_t*)ladr,len>80?80:len,16);
}

void dump_dpb(uint32_t dpbadr) {
    uint32_t da = dpbadr;
    uint32_t dd;
    printf("DPB at %08x:\r\n",da);
    while ((dd=*(uint32_t*)da)!=0) {
        printf("DSCPTR: %08x: %08x\r\n",da,dd);
        dump_dsc(dd);
        da += 4;
    }
}


void set_mpectl(int reg, uint32_t val) {
     *((uint32_t*)(MPECONTROL_BASEADDRESS+(reg*4))) = val;
    // printf("MPECTL: write of value 0x%08x to register %d complete.\r\n",val,reg);
}

void init_data() {
     debug_level = 0;
}

void circle(dim_t cx, dim_t cy, int r, void df(int x,int y)) {
    float phi, gx, gy;
    for (int i=0; i<MAXCIRCLE; i++) {
        phi = (i*2.0*PI)/(1.0*MAXCIRCLE);
        gx = cx+r*(0.0+(sin(phi*1.0)));
        gy = cy+r*(0.0+(cos(phi*1.0)));
        df(gx,gy);
    }
    printf("drawn circle\r\n");
}

void line(dim_t startx, dim_t starty, dim_t endx, dim_t endy, void df(int x,int y)) {
    // printf("starting line from %d/%d thru %d/%d.\n",startx,starty,endx,endy);
    int steps;
    float xi, yi;
    float dx = 1.0*(endx-startx);
    float dy = 1.0*(endy-starty);
    // printf("... line dx/dy = %f/%f.\n",dx,dy);
    steps = (abs(dx)>abs(dy))?abs(dx):abs(dy);
    // printf("... %d steps.\n",steps);
    xi = dx / (float) steps;
    yi = dy / (float) steps;
    // printf("... x/y increments are %f/%f.\n",xi,yi);
    float curx = (float)startx;
    float cury = (float)starty;
    for(int v=1; v < steps; v++) {
        df(curx,cury);
        curx += xi;
        cury += yi;
    }
    if (isDebug(DBG_DEBUG)) {
      printf("drawn line from %d/%d to %d/%d\r\n",startx,starty,endx,endy);
    }
}

void tick(float phi, void df(int x,int y)) {
    float sx = 120.0;
    float sy = 0;
    float ln = 7.0;
    struct Coord_t startpoint;
    struct Coord_t endpoint;
    struct Coord_t trans_start;
    struct Coord_t trans_end;
    startpoint.x = sx;
    startpoint.y = sy;
    endpoint.x = sx+ln;
    endpoint.y = sy;
    trans_start = translate(rotate(startpoint,phi),128.0,128.0);
    trans_end   = translate(rotate(endpoint,phi),128.0,128.0);
    line(trans_start.x,trans_start.y,trans_end.x,trans_end.y,df);
}


void ticks(void df(int x,int y)) {
    float phi;
    for (int i=0; i<12; i++) {
        phi = 1.0*i*PI/6.0;
        if (isDebug(DBG_DEBUG)) {
          printf("\r\n\r\n__ Tick Number %d with angle %9.7f :\r\n",i,phi);
        }
        tick(phi,df);
        setDebug(DBG_NONE);
    }
}

void hand(int htyp, int sixtieths, void df(int x,int y)) {
    float sx = 0.0;
    float sy = 0;
    float phi = ((60-(sixtieths))*2*PI)/60;
    float ln = htyp==0?110.0:htyp==1?90.0:70.0;
    struct Coord_t startpoint;
    struct Coord_t endpoint;
    struct Coord_t trans_start;
    struct Coord_t trans_end;
    startpoint.x = sx;
    startpoint.y = sy;
    endpoint.x = sx+ln;
    endpoint.y = sy;
    trans_start = translate(rotate(startpoint,phi),128.0,128.0);
    trans_end   = translate(rotate(endpoint,phi),128.0,128.0);
    line(trans_start.x,trans_start.y,trans_end.x,trans_end.y,df);
}

void clockface(void df(int x,int y)) {
    circle(128,128,128,df);
    ticks(df);
}

/*
//
// ================ class DPB - Descriptor Pointer Block =======================
//
typedef uint32_t dpb_t[DEF_DPBSIZE];   // a DPB is an array of pointer to DSCs, the DSC-pointers disassociated in 32-bit address form
typedef dpb_t* dpbp_t;                 // DPB-pointer
typedef struct dpbbuf_ts {             // a DPB buffer is a setup for working with a DPB
    dpbp_t dpbp;
    int capacity;
    int length;
} dpbbuf_t;

// allocate a new DBP with default size
dpbp_t new_dbp() {
    void* tt = malloc(DEF_DPBSIZE*sizeof(uint32_t));
    if (tt<=0) {
        printf("FATAL: Cannot allocate buffer of size %d\r\n",DEF_BUFSIZE);
        fatal();
    }
    return (dpbp_t) tt;
}


dpbbuf_t* new_dpbbuf() {
  void* bb = malloc(sizeof(dpb_t));
  if (bb<=0) {
    printf("FATAL: could not allocate DPB\r\n");
    fatal();
  }
  void* tt = malloc(sizeof(dpbbuf_t));
  if (tt<=0) {
    printf("FATAL: could not allocate DPBBUF\r\n");
        fatal();
    }
  dpbbuf_t* bp = (dpbbuf_t*)tt;
  bp->dpbp     = (dpbp_t)bb;
  bp->capacity = DEF_DPBSIZE;
  bp->length   = 0;
  (bp->dpbp)[0] = 0;             // terminate DPB at first position
  return bp;
}

void release_dpbbuf(dpbbuf_t* bp) {
    free((void*)bp->dpbp);
    free((void*)bp);
}

void dpbbuf_append_dscadr(dpbbuf_t* bp, uint32_t dscadr) {
    if (bp->length>=bp->capacity) {
      printf("FATAL: DPBBUF is full, no more DSC possible.\r\n");
      fatal();
    }
    (bp->dpbp)[bp->length++] = dscadr;       // apped DSCADR to and of current list and bump length
    (bp->dpbp)[bp->length] = 0;              // terminate list
}

uint32_t dpbbuf_get_dpbadr(dpbbuf_t* bp) {
    return (uint32_t)bp->dpbp;
}

*/

#define CCWIDTH  5
#define CCHEIGHT 7
#define CCXSTEP  7
#define CCYSTEP  9

#define CGBASE 0x20
#define CGTOP  0x3f

#define CGB(ccc) {(uint32_t)&ccc,((uint32_t)&ccc)+sizeof(ccc)-1}
#define CGBNULL  {0,0}

uint8_t char_blank[] = {0,0,0,0};                                                                         // <space>
uint8_t char_0[]     = {1,0,2,0, 3,0,4,1, 4,2,4,3, 4,4,4,5, 3,6,2,6, 1,6,0,5, 0,4,0,3, 0,2,0,1};          // 0
uint8_t char_1[]     = {1,0,3,0, 2,0,2,1, 2,2,2,3, 2,4,2,5, 2,6,1,5};                                     // 1
uint8_t char_2[]     = {4,0,3,0, 2,0,1,0, 0,0,0,1, 1,2,2,3, 3,4,4,5, 3,6,2,6, 1,6,0,5, 0,4,0,3};          // 2
uint8_t char_3[]     = {0,6,1,6, 2,6,3,6, 4,6,3,5, 2,4,3,4, 4,4,4,3, 4,2,4,1, 3,0,2,0, 1,0,0,1};          // 3
uint8_t char_4[]     = {0,6,0,5, 0,4,0,3, 1,3,2,3, 4,3,2,4, 2,3,2,2, 2,1,2,0};                            // 4
uint8_t char_5[]     = {4,6,3,6, 2,6,1,6, 0,6,0,5, 0,4,0,3, 0,2,1,3, 2,3,3,3, 4,2,4,1, 3,0,2,0, 1,0,0,1}; // 5
uint8_t char_6[]     = {4,5,3,6, 2,6,1,6, 0,5,0,4, 0,3,0,2, 0,1,1,0, 2,0,3,0, 4,1,4,2, 4,3,3,3, 2,3,1,3}; // 6
uint8_t char_7[]     = {0,6,1,6, 2,6,3,6, 4,6,3,5, 2,4,1,3, 0,2,0,1, 0,0,0,0};                            // 7
uint8_t char_8[]     = {0,1,0,2, 1,3,2,3, 3,3,4,4, 4,5,3,6, 2,6,1,6, 0,5,0,4, 4,2,4,1, 3,0,2,0, 1,0,0,1}; // 8
uint8_t char_9[]     = {0,1,1,0, 2,0,3,0, 4,1,4,2, 4,3,4,4, 4,5,3,6, 2,6,1,6, 0,5,0,4, 0,3,1,3, 2,3,3,3}; // 9
uint8_t char_colon[] = {2,1,3,1, 3,2,2,2, 2,4,3,4, 3,5,2,5};                                              // :
uint8_t char_point[] = {2,1,3,1, 3,2,2,2};                                                                // .
uint8_t char_dash[]  = {0,3,1,3, 2,3,3,3, 4,3,0,3};                                                       // -

uint32_t chargen[CGTOP+1-CGBASE][2] =
 {CGB(char_blank),CGBNULL,CGBNULL,CGBNULL,CGBNULL,CGBNULL,CGBNULL,CGBNULL,
  CGBNULL,CGBNULL,CGBNULL,CGBNULL,CGBNULL,CGB(char_dash),CGB(char_point),CGBNULL,
  CGB(char_0),CGB(char_1),CGB(char_2),CGB(char_3),CGB(char_4),CGB(char_5),CGB(char_6),CGB(char_7),
  CGB(char_8),CGB(char_9),CGB(char_colon),CGBNULL,CGBNULL,CGBNULL,CGBNULL,CGBNULL};

// put character <c> glyph into DBUF <bp>
void dbuf_set_char_cbuf(dbuf_t *bp, char c) {
    int cc = (int)c;
    if ((cc<CGBASE) || (cc>CGTOP)) {
      cc = CGBASE;
    }
    dbuf_set_cbufadr(bp,chargen[cc-CGBASE][0],chargen[cc-CGBASE][1]);
}


void reset_mpe() {
    led_turn_all_off(green_leds);
    set_mpectl(MPEREG_RUN,0);
    led_turn_on(green_leds,0);
    ioport_set_value(cport,0,0x80);              printf("NOS032 - set RESET to MPE\r\n");
    printf("NOS032_init - stopped MPE for 1000 ms\r\n");
    delay_ms(1000);
    printf("NOS032_init - startup delay done, continuing start\r\n");
    ioport_set_value(cport,0,0);                 printf("NOS032 - unset RESET to MPE\r\n");
    led_turn_on(green_leds,1);
}

// ============================ main initialization =====================================================

void init() {
    swplatform_init_stacks();
    init_data();
    reset_mpe();
}

// ============================ main program execution =====================================================

dbuf_t *gdbp;                        // global dbuf pointer for dcf_daw_func

void gdb_draw_func(int x, int y) {
    x = (x>255)?255:x;
    y = (y>255)?255:y;
    draw_to_dbuf(gdbp,x,y);
    if (isDebug(DBG_DEBUG)) {
      printf("NOS032 gdb_draw_func wrote X=%d, Y=%d at index=%d\r\n",x,y,gdbp->current_write_index-2);
    }
}

void nos_clock_single(int isDSC) {
    uint32_t dpb[2];
    dbuf_t *sgdbp = (dbuf_t*)0;
    int hours = 9;
    int minutes = 28;
    int seconds = 41;
    uint64_t tod;
    int curdip = ioport_get_value(drv_dip,0);
    int newdip = curdip;
    struct tm t;
    for (;;) {
      int ic = rtc_get_time(rtc_1,&t);
      if (ic!=0) {
        printf("NOS032.RTC-Error=%d.will read time from time_ms\r\n",ic);
        tod = clock_ms();
        tod /= 1000;
        seconds = tod % 60;
        tod /= 60;
        minutes = tod % 60;
        tod /= 60;
        hours = tod % 12;
      } else {
        seconds = t.tm_sec;
        minutes = t.tm_min;
        hours   = t.tm_hour;
        printf("NOS032.RTC-TIME=%d:%d:%d\r\n",hours,minutes,seconds);
        hours %= 12;
      }
      printf("NOS032_CLOCK.TIME=%d:%d:%d\r\n",hours,minutes,seconds);
      gdbp = new_dbuf();
      clockface(gdb_draw_func);
      hand(0,seconds,gdb_draw_func);
      hand(1,minutes,gdb_draw_func);
      hand(2,(hours*5)+(minutes/12),gdb_draw_func);
      led_turn_on(green_leds,2);
      // point MPlayer engine to the newly written buffer
      if (isDSC) {
        dpb[0] = dbuf_get_dscadr(gdbp);
        dpb[1] = 0;
        set_mpectl(MPEREG_ADR_DPB,(uint32_t)&dpb[0]);               // set the DPB base address
        printf("NOS032.NOT.SETUPDSCDPB.DPBADR=%08x\r\n",(uint32_t)&dpb[0]);
      } else {
        set_mpectl(MPEREG_ADR_LOW,gdbp->dsc.low_adr);
        set_mpectl(MPEREG_ADR_HIGH,gdbp->dsc.high_adr);
        printf("NOS032.CLOCK.SETUP_SINGLE_ADR.LOW=%08x.HIGH=%08x\r\n",gdbp->dsc.low_adr,gdbp->dsc.high_adr);
      }
      set_mpectl(MPEREG_SLOTLENGTH,100);                 // set the length of the timeslot
      if (isDSC) {
        set_mpectl(MPEREG_RUN,0xC0000000);                 // now start the player in DSC mode
        printf("NOS032.CLOCK.DSCMODE.RUN\r\n");
      } else {
        set_mpectl(MPEREG_RUN,0x80000000);                 // Now start the player in SINGLE buffer mode
        printf("NOS032.CLOCK.SINGLE.RUN\r\n");
      }
      led_turn_on(green_leds,3);
      delay_ms(400);
      if (sgdbp!=(dbuf_t*)0) {
        release_dbuf(sgdbp);
      }
      sgdbp = gdbp;
      newdip = ioport_get_value(drv_dip,0);
      if (curdip!=newdip) {
          break;
      }
      printf("NOS032.CLOCK.SINGLE.BREAK.OLDDIP=%d.NEWDIP=%d\r\n",curdip,newdip);
    }
}

// necessary DSC in the clock DPB:
//  1 for  face
//  1 for hands
//  8 for time display
//  1 for terminator
// ---------------------
// 11 total

#define CLKTIMDISPLEN 6
#define CLKDPBSIZE 11
// X-position of the time display
#define CLKTIMDISPX (MAXX-(CLKTIMDISPLEN*CCXSTEP))/2
// Y-position of the time display
#define CLKTIMDISPY 20

void dec2dig_dbuf(int num,dbuf_t* db1,dbuf_t* db2) {
    char ps[10];
    sprintf(ps,"%02d",num);
    dbuf_set_char_cbuf(db1,ps[0]);
    dbuf_set_char_cbuf(db2,ps[1]);
}

void nos_clock(int isDSC) {
  if (!isDSC) {
        nos_clock_single(isDSC);
        return;
  }
  dbuf_t *gdbp_face = (dbuf_t*)0;
  dbuf_t *gdbp_hands = (dbuf_t*)0;
  dbuf_t *gdbp_bin = (dbuf_t*)0;

    int hours = 9;
    int minutes = 28;
    int seconds = 41;
    uint64_t tod;
    struct tm t;
    init_data();
    uint32_t dpb[CLKDPBSIZE];
    dbuf_t* timedispdbufs[8];    // keep 8 dbufs for the time display
    for (int i=0; i<8; i++) {
      timedispdbufs[i] = new_dbuf_no_cbuf();
      dpb[i+2] = dbuf_get_dscadr(timedispdbufs[i]);
      dbuf_set_char_cbuf(timedispdbufs[i],((i==2)||(i==5))?':':' ');
      dbuf_set_offset(timedispdbufs[i],CLKTIMDISPX+(i*CCXSTEP),CLKTIMDISPY);
    }
    dpb[CLKDPBSIZE-1] = 0;      // terminate DPB
    gdbp = new_dbuf();
    clockface(gdb_draw_func);
    gdbp_face = gdbp;
    dpb[0] = dbuf_get_dscadr(gdbp_face);
  int ic;
  int curdip = ioport_get_value(drv_dip,0);
  int newdip = 0;
  for (;;) {
    ic = rtc_get_time(rtc_1,&t);
    if (ic!=0) {
      printf("NOS032.CLOCK_DSC.RTC-Error=%d.will read time from time_ms\r\n",ic);
      tod = clock_ms();
      tod /= 1000;
      seconds = tod % 60;
      tod /= 60;
      minutes = tod % 60;
      tod /= 60;
      hours = tod % 12;
    } else {
      seconds = t.tm_sec;
      minutes = t.tm_min;
      hours   = t.tm_hour;
      printf("NOS032.CLOCK_DSC.RTC-TIME=%d:%d:%d    \r",hours,minutes,seconds);
    }
    led_turn_on(green_leds,2);
    // paint the hands
    gdbp = new_dbuf();                                  // open a new dbuf for the current hands
    hand(0,seconds,gdb_draw_func);
    hand(1,minutes,gdb_draw_func);
    hand(2,(hours%12)*5+minutes/12,gdb_draw_func);
    dpb[1] = dbuf_get_dscadr(gdbp);                     // update hands entry in DPB
    gdbp_hands = gdbp;
    if (isDebug(DBG_DEBUG)) {
      dump_dpb((uint32_t)&dpb[0]);
    }
    // set the time display (prepared in initialization above)
    dec2dig_dbuf(hours,  timedispdbufs[0],timedispdbufs[1]);
    dec2dig_dbuf(minutes,timedispdbufs[3],timedispdbufs[4]);
    dec2dig_dbuf(seconds,timedispdbufs[6],timedispdbufs[7]);
    set_mpectl(MPEREG_ADR_DPB,(uint32_t)&dpb[0]);               // set the DPB base address
    // printf("NOS032.SETUPDSCDPB.DPBADR=%08x\r\n",(uint32_t)&dpb[0]);
    set_mpectl(MPEREG_SLOTLENGTH,200);                 // set the length of the timeslot 1/50MHz(20ns)*SLOTLENGTH
    set_mpectl(MPEREG_RUN,0xE0000000);                 // now (re)start the player in DSC-EXTENDED mode
    // printf("NOS032.CLOCK.DSCMODE.RUN\r\n");
    led_turn_on(green_leds,3);
    delay_ms(1000);
    if (gdbp_bin!=(dbuf_t*)0) {                        // throw away old dbuf from waste bin (if any in there)
        release_dbuf(gdbp_bin);
      }
    gdbp_bin = gdbp_hands;                             // put current hands dbuf in waste bin
    newdip = ioport_get_value(drv_dip,0);
    if (newdip!=curdip) {
        break;
    }
  }
  printf("\r\nNOS032.CLOCK.DSCMODE.BREAK.OLDDIP=%d.NEWDIP=%d\r\n",curdip,newdip);
}

void wbctest() {
    uint32_t curdip, newdip;
    printf("NOS032.WBCTEST.START\r\n");
    curdip = ioport_get_value(drv_dip,0);
    for (;;) {
        set_mpectl(MPEREG_ADR_LOW,0);
        set_mpectl(MPEREG_ADR_HIGH,4);
        set_mpectl(MPEREG_RUN,0xC0000000);
        newdip = ioport_get_value(drv_dip,0);
        if (newdip!=curdip) break;
    }
    printf("NOS032.WBCTEST.END\r\n");
}

void nos_dsctest() {
    uint32_t dpb[2];
    init_data();
    gdbp = new_dbuf();
    printf("NOS032.DSCTEST.START\r\n");
    gdb_draw_func(0,0);
    gdb_draw_func(255,0);
    gdb_draw_func(255,255);
    gdb_draw_func(0,255);
    dpb[0] = dbuf_get_dscadr(gdbp);
    dpb[1] = 0;
    printf("NOS032.DSCTEST.SETUP\r\n");
    dump_dpb((uint32_t) &dpb[0]);
    // ... and now set the MPE
    set_mpectl(MPEREG_SLOTLENGTH,100);
    set_mpectl(MPEREG_ADR_DPB,(uint32_t) &dpb[0]);
    set_mpectl(MPEREG_RUN,0xC0000000);                            // set RUN and DSCMODE
    printf("NOS032.DSCTEST.RUN\r\n");
}

void nos_csqtest() {
    printf("NOS032.CSQTEST.START\r\n");
    printf("NOS032.CSQTEST.PRE_START_DELAY\r\n");
    // wait for a rather long time ...
    delay_ms(2000);
    // ... and now set the MPE
    set_mpectl(MPEREG_RUN,0x81000001);                            // set RUN and OPMODE=1, counter = 1
    printf("NOS032.CSQTEST.RUN\r\n");
}

/*
void nos_memtest() {
    uint32_t adr,dat,rdat;
    uint32_t lowmem = (uint32_t)&cbuf;
    uint32_t highmem = (uint32_t)&cbuf+sizeof(cbuf);
    dat = 0x00000001;
    printf("NOS032 memtest, dscs   is at %08x and has size of %08x\r\n",(uint32_t)&dscs,sizeof(dscs));
    printf("NOS032 memtest, dpbs   is at %08x and has size of %08x\r\n",(uint32_t)&dpbs,sizeof(dpbs));
    printf("NOS032 memtest, cbuf   is at %08x and has size of %08x\r\n",(uint32_t)&cbuf,sizeof(cbuf));
    printf("NOS032 memtest, upmark is at %08x and has size of %08x\r\n",(uint32_t)&upmark,sizeof(upmark));
    printf("NOS032 memory test between %08x and %08x\r\n",lowmem,highmem);
    printf("writing: ");
    for (adr=lowmem;adr<highmem;adr+=4) {
        if ((adr%256)==0) {
            printf(".");
        }
        *((uint32_t*)adr) = dat;
        if (dat==0x10000000) {
            dat = 1;
        } else {
          dat = dat << 1;
        }
    }
    dat = 1;
    printf("reading: ");
    for (adr=lowmem;adr<highmem;adr+=4) {
       if ((adr%256)==0) {
            printf(".");
        }
      rdat = *((uint32_t*)adr);
      if (rdat!=dat) {
        printf("NO0S32 memory verification error, address = %08x, expected = %08x, read = %08x.\r\n",adr,dat,rdat);
        break;
      }
      if (dat==0x10000000) {
            dat = 1;
      } else {
          dat = dat << 1;
      }
    }
    printf("\r\nNOS032 memory test complete.\r\n");
}
*/
void main() {
    init();
    uint64_t fhz = freq_hz();
    printf("NOS032008005, build %s, %s started, clock is %d Hz\r\n",__TIME__, __DATE__,freq_hz());
    uint32_t curdip = OPERMODE;
    printf("NOS032 default operation mode is %d\r\n",curdip);
    uint32_t newdip;
    for (;;) {
      reset_mpe();
      set_mpectl(MPEREG_RUN,0);                 // reset MPEngine
      curdip = ioport_get_value(drv_dip,0);
      printf("NOS032 DIP switch setting is %d\r\n",curdip);
      switch (curdip) {
      case (OPM_CLOCK):
      case (OPM_CLOCK_DSC):
         nos_clock(curdip==OPM_CLOCK_DSC);
         break;
      case (OPM_DSCTEST):
         nos_dsctest();
         break;
      case (OPM_WBCTEST):
         wbctest();
         break;
      case (OPM_CSQTEST):
         nos_csqtest();
         break;
      /*
      case (OPM_MEMTEST):
         nos_memtest();
         break;
      */
      default:
        printf("NOS032 unknown operation mode selected, reset DIP switch and restart\r\n");
      }
      for (;;) {
        newdip = ioport_get_value(drv_dip,0);
        if (newdip!=curdip) break;
      }
    }
}

